package de.uniba.kinf.geogames.esriguesstimate.activities;

import android.content.ComponentName;
import android.content.Context;
import android.content.Intent;
import android.content.ServiceConnection;
import android.graphics.Color;
import android.location.Location;
import android.os.Bundle;
import android.os.IBinder;
import android.support.v7.app.ActionBar;
import android.support.v7.app.ActionBarActivity;
import android.util.Log;
import android.view.Menu;
import android.view.MenuItem;
import android.view.View;
import android.view.ViewGroup;
import android.widget.Button;
import android.widget.TextView;
import android.widget.Toast;

import com.esri.android.map.GraphicsLayer;
import com.esri.android.map.Layer;
import com.esri.android.map.MapView;
import com.esri.android.map.event.OnStatusChangedListener;
import com.esri.core.geometry.GeometryEngine;
import com.esri.core.geometry.LinearUnit;
import com.esri.core.geometry.Point;
import com.esri.core.geometry.Polygon;
import com.esri.core.geometry.SpatialReference;
import com.esri.core.geometry.Unit;
import com.esri.core.map.Graphic;
import com.esri.core.portal.WebMap;
import com.esri.core.symbol.SimpleLineSymbol;
import com.esri.core.symbol.SimpleMarkerSymbol;
import com.esri.core.symbol.Symbol;
import com.esri.core.symbol.TextSymbol;

import java.util.ArrayList;

import de.uniba.kinf.geogames.esriguesstimate.R;
import de.uniba.kinf.geogames.esriguesstimate.data.GeoLocation;
import de.uniba.kinf.geogames.esriguesstimate.data.GeoPosition;

/**
 * Created by Thomas on 19.03.2015.
 */
public class GameMap extends ActionBarActivity implements GameService.GameServiceListener {

    protected static final String TAG = "GameMap";
    public static final double GUESSING_CORRECTION_FACTOR = 2.0;

    protected static SpatialReference mapSpRef;

    protected GameService gameService;
    protected boolean gameServiceBound;

    // -- UI --
    protected ActionBar actionBar;
    protected TextView scoreText;
    protected TextView timeText;
    protected MapView map;
    protected boolean mapInit = false;
    protected Button actionBtn;
    protected Button timerBtn;
    protected GraphicsLayer graphicsLayer;
    protected Graphic playerPosGraphic;
    protected int playerPosGraphicId;
    protected Graphic targetPosGraphic;
    protected int targetPosGraphicId;
    protected ArrayList<Integer> helperGraphics = new ArrayList<Integer>();

    // -- Game Play --
    protected String mapName = "";
    protected int playerScore = 0;
    protected int remainingSeconds = 0;

    // ===== Activity Methods =====

    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);
        // Init UI
        setContentView(R.layout.activity_map);
        actionBar = getSupportActionBar();
        actionBtn = (Button) findViewById(R.id.mapActionBtn);
        actionBtn.setEnabled(false);
        actionBtn.setClickable(false);
        timerBtn = (Button) findViewById(R.id.mapTimerBtn);
        timerBtn.setOnClickListener(new View.OnClickListener() {
            @Override
            public void onClick(View v) {
                disableGuesses();
            }
        });
        scoreText = (TextView) findViewById(R.id.scoreText);
        timeText = (TextView) findViewById(R.id.timeText);

        // Bind to service
        Intent intent = new Intent(this, GameService.class);
        bindService(intent, gameServiceCon, Context.BIND_AUTO_CREATE);
        Log.d(TAG, "started");
    }

    @Override
    protected void onDestroy() {
        if(gameServiceBound) {
            gameService.unregisterListener(GameMap.this);
            unbindService(gameServiceCon);
            gameServiceBound = false;
        }
        stopService(new Intent(this, GameService.class));
        super.onDestroy();
    }

    @Override
    public void onBackPressed() {
    }

    @Override
    public boolean onCreateOptionsMenu(Menu menu) {
        getMenuInflater().inflate(R.menu.menu_map, menu);
        return true;
    }

    @Override
    public boolean onOptionsItemSelected(MenuItem item) {
        int id = item.getItemId();
        if (id == R.id.action_settings) {
            disableGuesses();
            gameService.onPlayerReachedTarget(false);
        }
        return super.onOptionsItemSelected(item);
    }

    // ===== Game Service Connection =====

    private ServiceConnection gameServiceCon = new ServiceConnection() {
        @Override
        public void onServiceConnected(ComponentName className, IBinder service) {
            GameService.LocalBinder binder = (GameService.LocalBinder) service;
            gameService = binder.getService();
            gameServiceBound = true;
            gameService.registerListener(GameMap.this);
            Log.d(TAG, "Service bound");

            WebMap webmap = gameService.getWebmap();
            actionBar.setTitle("Guesstimate - " + gameService.getMapName());
            actionBar.setSubtitle("Please wait. Loading...");
            scoreText.setText("Game duration: " + gameService.getGameDuration().text);
            init(webmap);

            ((ViewGroup) findViewById(R.id.playMapContainer)).addView(map,
                    new ViewGroup.LayoutParams(ViewGroup.LayoutParams.MATCH_PARENT,
                            ViewGroup.LayoutParams.MATCH_PARENT));

            updatePlayerPosition();
        }
        @Override
        public void onServiceDisconnected(ComponentName arg0) {
        }
    };

    private void init(final WebMap webmap) {
        mapName = webmap.getInfo().getTitle();

        map = new MapView(getApplicationContext(), webmap, null, null);
        map.setOnStatusChangedListener(new OnStatusChangedListener() {
            public void onStatusChanged(Object source, STATUS status) {
                if (mapInit == false && OnStatusChangedListener.STATUS.INITIALIZED == status && source == map) {
                    Log.d(TAG, "Initializing map view");
                    mapInit = true;
                    // Pass feature service to game service
                    mapSpRef = map.getSpatialReference();
                    Log.d(TAG, "Map layers: " + map.getLayers().length);
                    String featureServiceUrl = "";
                    for (Layer layer : map.getLayers()) {
                        Log.d(TAG, "Layer Title: " + layer.getTitle());
                        Log.d(TAG, "Layer Name: " + layer.getName());
                        if (layer.isWebMapBaselayer() == false) {
                            layer.setVisible(false);
                            featureServiceUrl = layer.getUrl();
                            featureServiceUrl = featureServiceUrl.substring(0, featureServiceUrl.length() - 1);
                            break;
                        }
                    }
                    // Add graphics layer
                    graphicsLayer = new GraphicsLayer();
                    map.addLayer(graphicsLayer);
                    gameService.initFeatureServiceLayer(featureServiceUrl, true);
                }
            }
        });
    }

    private void startPickActivity() {
        Log.d(TAG, "Starting 'guessing' activity reached");
        Intent pickIntent = new Intent(getBaseContext(), BlindGuess.class);
        startActivity(pickIntent);
    }


    private void startGameEndActivity() {
        Log.d(TAG, "Starting 'game over' activity reached");
        Intent pickIntent = new Intent(getBaseContext(), GameOver.class);
        startActivity(pickIntent);
    }

    // ===== Game Service Listener Methods =====

    @Override
    public void updateRemainingSeconds(int seconds) {
        remainingSeconds = seconds;
        updatePlayerStats();
    }

    @Override
    public void updatePlayerScore(int score, int deltaScore) {
        Log.d(TAG, "Player score updated ("+score+")");
        playerScore = score;
        Toast.makeText(getApplicationContext(), "Scored "+deltaScore+" points!",
                Toast.LENGTH_LONG).show();
        updatePlayerStats();
    }

    @Override
    public void updatePlayerPosition() {
        Log.d(TAG, "Updating player position");
        if(playerPosGraphic != null) {
            graphicsLayer.removeGraphic(playerPosGraphicId);
        }
        if(graphicsLayer != null) {
            Location playerPos = gameService.getPlayerPosition();
            if(playerPos == null) {
                Log.d(TAG, "Player position not available");
                return;
            }
            Log.d(TAG, "Pos: "+playerPos.getLatitude()+","+playerPos.getLongitude());
            final Point p = new Point(playerPos.getLongitude(), playerPos.getLatitude());

            final Symbol symbol = new SimpleMarkerSymbol(Color.GREEN, 15, SimpleMarkerSymbol.STYLE.CIRCLE);
            final Point pg = (Point) GeometryEngine.project(p, SpatialReference.create(4326), map.getSpatialReference()); // TODO: Test SpatialReference.WGS84
            playerPosGraphic = new Graphic(pg, symbol);

            playerPosGraphicId = graphicsLayer.addGraphic(playerPosGraphic);
            Log.d(TAG, "" + pg.getX());
        }
    }

    @Override
    public void updateTargetPosition() {
        Log.d(TAG, "Updating target position");
        timerBtn.setVisibility(View.VISIBLE);

        if(targetPosGraphic != null) {
            graphicsLayer.removeGraphic(targetPosGraphicId);
        }
        if(graphicsLayer != null) {
            GeoPosition targetPos = gameService.getTargetPos();
            if(targetPos != null) {
                final Point p = new Point(targetPos.getLng(), targetPos.getLat());
                Log.d(TAG, "New target pos: " + targetPos.getLat() + "," + targetPos.getLng());
                final Symbol symbol = new SimpleMarkerSymbol(Color.RED, 25, SimpleMarkerSymbol.STYLE.X);
                final Point pg = (Point) GeometryEngine.project(p, SpatialReference.create(4326), map.getSpatialReference());
                targetPosGraphic = new Graphic(pg, symbol);
                targetPosGraphicId = graphicsLayer.addGraphic(targetPosGraphic);
                map.centerAt(pg, true);

                for(GeoLocation guess: gameService.getLocationGuesses()) {
                    final Point gp = new Point(guess.getLng(), guess.getLat());
                    final Point gpg = (Point) GeometryEngine.project(gp, SpatialReference.create(4326), map.getSpatialReference());

                    Polygon polygon = GeometryEngine.buffer(gpg, map.getSpatialReference(), guess.getDistanceGuess() * GUESSING_CORRECTION_FACTOR, Unit.create(LinearUnit.Code.METER));
                    SimpleLineSymbol s1 = new SimpleLineSymbol(Color.RED, 1f);
                    TextSymbol txtSymbol = new TextSymbol(15, guess.getName(), Color.BLACK);
                    Graphic g = new Graphic(polygon, s1);
                    Graphic h = new Graphic(polygon, txtSymbol);
                    helperGraphics.add(graphicsLayer.addGraphic(g));
                    helperGraphics.add(graphicsLayer.addGraphic(h));

                    final Symbol guessSymbol = new SimpleMarkerSymbol(Color.BLACK, 10, SimpleMarkerSymbol.STYLE.CIRCLE);
                    final Graphic guessGraphic = targetPosGraphic = new Graphic(gpg, guessSymbol);
                    helperGraphics.add(graphicsLayer.addGraphic(guessGraphic));

                }
            }
        }
    }


    @Override
    public void playerReachedTarget(boolean gameStart) {
        if(gameStart) {
            Log.d(TAG, "Player at target (games start)");
            showStartGameButton();
        } else {
            Log.d(TAG, "Player at target");
            showGuessingButton();
        }
    }

    @Override
    public void gameOver() {
        startGameEndActivity();
    }

    // ===== Private Methods =====

    private void updatePlayerStats() {
        scoreText.setText("Score: "+playerScore);
        timeText.setText("" + remainingSeconds + " seconds remaining");
    }

    private void showStartGameButton() {
        runOnUiThread(new Runnable() {
            public void run() {
                Log.d(TAG, "Enabling start game button");
                actionBtn.setText(getResources().getString(R.string.start_game));

                actionBtn.setOnClickListener(new View.OnClickListener() {
                    @Override
                    public void onClick(View v) {
                        hideButton();
                        gameService.startCountdown();
                        startPickActivity();

                        actionBar.setSubtitle("Move to the target destination.");
                    }
                });
                actionBtn.setClickable(true);
                actionBtn.setEnabled(true);
                actionBtn.setVisibility(View.VISIBLE);

                actionBar.setSubtitle("Press \"Start Game\" to begin.");
            }
        });
    }

    private void showGuessingButton() {
        runOnUiThread(new Runnable() {
            public void run() {
                Log.d(TAG, "Enabling guessing button");
                actionBtn.setText(getResources().getString(R.string.start_guessing));
                actionBtn.setOnClickListener(new View.OnClickListener() {
                    @Override
                    public void onClick(View v) {
                        hideButton();
                        startPickActivity();

                        actionBar.setSubtitle("Move to the destination.");
                    }
                });
                actionBtn.setClickable(true);
                actionBtn.setEnabled(true);
                actionBtn.setVisibility(View.VISIBLE);

                actionBar.setSubtitle("Make your guesses, when ready.");
            }
        });
    }

    private void hideButton() {
        runOnUiThread(new Runnable() {
            public void run() {
                actionBtn.setOnClickListener(null);
                actionBtn.setClickable(false);
                actionBtn.setVisibility(View.INVISIBLE);
            }
        });
    }

    private void disableGuesses() {
        timerBtn.setVisibility(View.INVISIBLE);
        for(Integer i : helperGraphics) {
            graphicsLayer.removeGraphic(i);
        }
    }

}
