package de.uniba.kinf.geogames.esriguesstimate.activities;

import android.app.PendingIntent;
import android.app.Service;
import android.content.Context;
import android.content.Intent;
import android.location.Location;
import android.location.LocationListener;
import android.location.LocationManager;
import android.os.Binder;
import android.os.Bundle;
import android.os.CountDownTimer;
import android.os.IBinder;
import android.support.v4.app.NotificationCompat;
import android.util.Log;

import com.esri.core.geodatabase.GeodatabaseFeature;
import com.esri.core.geodatabase.GeodatabaseFeatureServiceTable;
import com.esri.core.geometry.GeometryEngine;
import com.esri.core.geometry.Point;
import com.esri.core.geometry.SpatialReference;
import com.esri.core.map.CallbackListener;
import com.esri.core.map.FeatureResult;
import com.esri.core.portal.Portal;
import com.esri.core.portal.WebMap;
import com.esri.core.tasks.query.QueryParameters;

import java.util.ArrayList;
import java.util.Random;

import de.uniba.kinf.geogames.esriguesstimate.R;
import de.uniba.kinf.geogames.esriguesstimate.calculations.EstimationCalculator;
import de.uniba.kinf.geogames.esriguesstimate.calculations.PositionCalculator;
import de.uniba.kinf.geogames.esriguesstimate.data.GeoLocation;
import de.uniba.kinf.geogames.esriguesstimate.data.GeoPosition;
import de.uniba.kinf.geogames.esriguesstimate.game.GameDuration;

/**
 * Created by Thomas on 18.03.2015.
 */
public class GameService extends Service {

    protected static final String TAG = "GameService";

    private static final long MINIMUM_DISTANCECHANGE_FOR_UPDATE = 1; // in Meters
    private static final long MINIMUM_TIME_BETWEEN_UPDATE = 1000; // in Milliseconds

    public static final String FIELD_ID_IDENTIFIER = "FIELD_ID";
    public static final int MAX_LOCATIONS_TO_PICK = 4;
    public static final int REQUIRED_GUESSES = 4;
    public static final float PROXY_RADIUS = 40.0f;

    private final IBinder binder = new LocalBinder();
    private final ArrayList<GameServiceListener> listeners = new ArrayList<GameServiceListener>();

    // -- Gameplay --
    private String gameFieldId;
    private GameDuration gameDuration;
    private CountDownTimer timer;

    private boolean gameHasStarted = false;
    private boolean playerHasTarget = false;
    private Location playerPosition = null;
    private GeoPosition targetPosition = null;

    private int playerScore = 0;

    protected LocationManager locService;
    protected LocationListener locListener;

    protected WebMap webmap;
    protected String mapName = "";
    protected GeodatabaseFeatureServiceTable featureService;
    protected ArrayList<GeoLocation> locationsToPick = new ArrayList<>();
    protected ArrayList<GeoLocation> locGuesses = new ArrayList<>();
    protected final PositionCalculator calculator = new EstimationCalculator();

    // ===== Service Methods =====

    @Override
    public int onStartCommand(Intent intent, int flags, int startId) {
        // Get game settings
        Bundle dataBundle = intent.getExtras();
        gameFieldId = dataBundle.getString(FIELD_ID_IDENTIFIER);
        gameDuration = GameDuration.fromValue(dataBundle.getInt(GameDuration.IDENTIFIER));

        initLocationService();
        initWebMap();
        initNotification();

        Log.d(TAG, "started");
        return Service.START_NOT_STICKY;
    }

    @Override
    public IBinder onBind(Intent intent) {
        Log.d(TAG, "bound");
        return binder;
    }

    @Override
    public void onDestroy() {
        super.onDestroy();
    }

    // ===== Game Service Methods =====

    public void startCountdown() {
        int seconds = gameDuration.getSeconds();
        Log.d(TAG, "Countdown started: "+seconds+" remaining");
        timer = new CountDownTimer(seconds * 1000 , 1000) {
            @Override
            public void onTick(long millisUntilFinished) {
                for(GameServiceListener listener : listeners) {
                    listener.updateRemainingSeconds((int)(millisUntilFinished/1000));
                }
            }
            @Override
            public void onFinish() {
                Log.d(TAG, "Countdown finished");
                for(GameServiceListener listener : listeners) {
                    listener.gameOver();
                }
            }
        };
        timer.start();
        gameHasStarted = true;
    }

    public void addGuess(GeoLocation guess) {
        Log.d(TAG, "Guess added: " + guess.getName() + " - " + guess.getDistanceGuess());
//        for(GeoLocation loc: locGuesses) {
//            if(loc.getName().equals(guess.getName())) {
//                locGuesses.remove(loc);
//            }
//        }
        locGuesses.add(guess);
        if(locGuesses.size() >= REQUIRED_GUESSES) {
            GeoPosition.init(locGuesses.get(0).getLat(), locGuesses.get(0).getLng());
            addProximityAlert(calculator.calcGeoPosition(locGuesses));
        }
    }

    public void removeGuess(GeoLocation guess) {
        Log.d(TAG, "Guess removed: " + guess.getName());
        locGuesses.remove(guess);
    }

    // -- Initialization --

    public void initFeatureServiceLayer(String url, final boolean guess) {
        Log.d(TAG, "Initializing feature service initialized");
        if(EsriLogin.isLoggedIn()) {
            this.featureService = new GeodatabaseFeatureServiceTable(url, EsriLogin.getCredentials(), 0);
        } else {
            this.featureService = new GeodatabaseFeatureServiceTable(url, 0);
        }
        featureService.setFeatureRequestMode(GeodatabaseFeatureServiceTable.FeatureRequestMode.ON_INTERACTION_NO_CACHE);
        featureService.initialize(new CallbackListener<GeodatabaseFeatureServiceTable.Status>() {
            @Override
            public void onCallback(GeodatabaseFeatureServiceTable.Status status) {
                if(status.equals(GeodatabaseFeatureServiceTable.Status.INITIALIZED)) {
                    Log.d(TAG, "Feature service initialized");
                    if (guess) {
                        onPlayerReachedTarget(true);
                    }
                }
            }
            @Override
            public void onError(Throwable throwable) {
                Log.d(TAG, "Error on setting feature table");
                throwable.printStackTrace();
            }
        });
    }

    // -- Getter & Setter --

    public GameDuration getGameDuration() {
        return gameDuration;
    }

    public WebMap getWebmap() {
        return webmap;
    }

    public String getMapName() { return mapName; }

    public boolean gameHasStarted() {
        return gameHasStarted;
    }

    public Location getPlayerPosition() {
        return playerPosition;
    }

    public boolean playerHasTarget() {
        return playerHasTarget;
    }

    public GeoPosition getTargetPos() {
        return targetPosition;
    }

    public ArrayList<GeoLocation> getLocationsToPick() {
        return locationsToPick;
    }

    public ArrayList<GeoLocation> getLocationGuesses() { return locGuesses; }

    public int getPlayerScore() {
        return this.playerScore;
    }

    // ===== Private Methods =====

    private void initLocationService() {
        Log.d(TAG, "Location changed");
        locService = (LocationManager) getSystemService(Context.LOCATION_SERVICE);
        this.locListener = new LocationListener() {
            @Override
            public void onLocationChanged(Location location) {
                Log.d(TAG, "Location changed");
                playerPosition = location;
                // Alert listeners about changed player position
                for(GameServiceListener listener : listeners) {
                    listener.updatePlayerPosition();
                }
                // Alert listeners if player is near target
                if(targetPosition != null) {
                    Location targetLocation = new Location("");
                    targetLocation.setLatitude(targetPosition.getLat());
                    targetLocation.setLongitude(targetPosition.getLng());
                    float distance = playerPosition.distanceTo(targetLocation);
                    Log.d(TAG, "Distance to target: "+distance);
                    if (distance < PROXY_RADIUS) {
                        onPlayerReachedTarget(false);
                    }
                }
            }
            @Override
            public void onStatusChanged(String provider, int status, Bundle extras) {}
            @Override
            public void onProviderEnabled(String provider) {}
            @Override
            public void onProviderDisabled(String provider) {}
        };
        locService.requestLocationUpdates(LocationManager.GPS_PROVIDER,
                MINIMUM_TIME_BETWEEN_UPDATE, MINIMUM_DISTANCECHANGE_FOR_UPDATE, locListener);

    }

    private void initWebMap() {
        Log.d(TAG, "Catching webmap");
        Portal portal = EsriLogin.getPortal(this);
        try {
            Log.d(TAG, "Webmap loaded");
            this.webmap = WebMap.newInstance(gameFieldId, portal);
            this.mapName = webmap.getInfo().getTitle();
            webmap.getOperationalLayers().get(0).getUrl();


        } catch (Exception e) {
            Log.d(TAG, "Error while catching webmap", e);
        }
    }

    private void initNotification() {

        Intent i = new Intent(this, GameMap.class);
        i.setFlags(Intent.FLAG_ACTIVITY_CLEAR_TOP | Intent.FLAG_ACTIVITY_SINGLE_TOP);
        PendingIntent pi = PendingIntent.getActivity(this, 0, i, 0);

        String noteText = getResources().getString(R.string.note_text)+" "+mapName;

        NotificationCompat.Builder mBuilder =
                new NotificationCompat.Builder(this)
                        .setSmallIcon(android.R.drawable.ic_dialog_map)
                        .setContentTitle(getResources().getString(R.string.note_title))
                        .setContentText(noteText)
                        .setContentIntent(pi);

        int mNotificationId = 001;
        startForeground(mNotificationId, mBuilder.build());
    }

    public void onPlayerReachedTarget(boolean gameStart) {
        Log.d(TAG, "Player has reached target");
        if(!gameStart) {
            removeProximityAlert();
            playerScore = playerScore + 100;
            for(GameServiceListener listener : listeners) {
                listener.updatePlayerScore(playerScore, 100);
            }
        }
        notifyListenersPlayerTargetReached(gameStart);
    }

    private void notifyListenersPlayerTargetReached(final boolean gameStart) {
        Log.d(TAG, "Quering feature table for locations");
        locationsToPick.clear();
        locGuesses.clear();
        QueryParameters qp = new QueryParameters();
        qp.setReturnGeometry(true);
        //qp.setMaxFeatures(MAX_LOCATIONS_TO_PICK);
        qp.setWhere("OBJECTID > 0");
        featureService.queryFeatures(qp, new CallbackListener<FeatureResult>() {
            @Override
            public void onCallback(FeatureResult objects) {

                int maxImportance = 0;
                ArrayList<GeoLocation> gameFieldLocations = new ArrayList<GeoLocation>();

                for (Object objFeature : objects) {
                    GeodatabaseFeature gf = (GeodatabaseFeature) objFeature;
                    Point p = (Point) gf.getGeometry();
                    p = (Point) GeometryEngine.project(p, GameMap.mapSpRef,
                            SpatialReference.create(SpatialReference.WKID_WGS84));
                    String name = (String) gf.getAttributeValue("name");
                    String desc = (String) gf.getAttributeValue("desc");
                    int importance = 1;
                    try {
                        importance = (int) gf.getAttributeValue("importance");
                        importance = Math.max(1, importance);
                    } catch (Exception e) {
                        Log.d(TAG, "Could not read importance value for "+ name);
                    }
                    Log.d(TAG, "Added feature to pick list: " + name + " " + p.getX() + "/" + p.getY());
                    GeoLocation geoLoc = new GeoLocation(name, desc, p.getY(), p.getX());
                    geoLoc.setImportance(importance);
                    gameFieldLocations.add(geoLoc);
                    maxImportance += importance;
                }

                for (int i = 0; i < MAX_LOCATIONS_TO_PICK; i++) {
                    Random r = new Random();
                    int pick = r.nextInt(maxImportance);
                    for(GeoLocation loc: gameFieldLocations) {
                        pick -= loc.getImportance();
                        if(pick <= 0) {
                            locationsToPick.add(loc);
                            maxImportance -= loc.getImportance();
                            gameFieldLocations.remove(loc);
                            break;

                        }
                    }
                }

                // Notify listeners
                for (GameServiceListener listener : listeners) {
                    listener.playerReachedTarget(gameStart);
                }
            }
            @Override
            public void onError(Throwable throwable) {
                Log.d(TAG, "Error on querying feature service");
                throwable.printStackTrace();
            }
        });

    }

    private void addProximityAlert(GeoPosition position) {
        targetPosition = position;
        playerHasTarget = true;
        Log.d(TAG, "Proximity alert added: " + position.getLat() + "," + position.getLng());
        for(GameServiceListener listener : listeners) {
            listener.updateTargetPosition();
        }
    }

    private void removeProximityAlert() {
        targetPosition = null;
        playerHasTarget = false;
        Log.d(TAG, "Proximity alert removed");
    }



    // ===== Game Service Listener Methods =====

    public void registerListener(GameServiceListener listener) {
        listeners.add(listener);
    }

    public void unregisterListener(GameServiceListener listener) {
        listeners.remove(listener);
    }

    public interface GameServiceListener {

        public void updateRemainingSeconds(int seconds);

        public void updatePlayerScore(int score, int deltaScore);

        public void updatePlayerPosition();

        public void updateTargetPosition();

        public void playerReachedTarget(boolean startGame);

        public void gameOver();

    }

    // ===== Local Binder =====

    public class LocalBinder extends Binder {
        GameService getService() {
            return GameService.this;
        }
    }

}
