package de.uniba.kinf.geogames.esriguesstimate.calculations;

import org.ejml.simple.SimpleMatrix;

import java.util.ArrayList;
import java.util.List;

import de.uniba.kinf.geogames.esriguesstimate.data.GeoLocation;
import de.uniba.kinf.geogames.esriguesstimate.data.GeoPosition;


/**
 * A PositionCalculator that calculates the target position by using circular
 * lateration.
 * 
 * @ see PositionCalculator
 * 
 * @author Thomas Heinz
 * 
 */
public class EstimationCalculator implements PositionCalculator {

    protected static final String LOGTAG = "EstimationCalculator";

    /**
     * Maximal derivation from the target distances from the given distances.
     * Break condition for the iterative calculating algorithm.
     */
    private static final double MAX_DEVIATION = 0.001;
    /**
     * If the algorithm should not terminate because the estimations does not
     * get better it stops after the maximal amount of iterations stated here
     * and returns the last guess.
     */
    private static final double MAX_ITERATIONS = 4;


    public GeoPosition calcGeoPosition(List<GeoLocation> guessList) {
        ArrayList<GeoPosition> posList = new ArrayList<GeoPosition>();
        ArrayList<Integer> distList  = new ArrayList<Integer>();
        for(GeoLocation guess:guessList) {
            if(guess.hasGuess()) {
                posList.add(GeoPosition.fromGeoLocation(guess));
                distList.add(guess.getDistanceGuess());
            }
        }
        return calcLocation(posList, distList);
    }

    /**
     * {@inheritDoc}
     *
     */
    public GeoPosition calcLocation(ArrayList<GeoPosition> posList,
                                    ArrayList<Integer> distList) {
        return calcLocation(posList, distList, new ArrayList<String>());
    }


    public GeoPosition calcLocation(ArrayList<GeoPosition> posList,
                                    ArrayList<Integer> distList, ArrayList<String> geoJson) {
        System.out.println("Calculating estimated position (max derivation: "
                + MAX_DEVIATION + ")");

        double error = Double.MAX_VALUE;
        int iterationIndex = 0;

        GeoPosition estimation = getBoxCenter(posList);

        // Create empty calculation matrixes
        SimpleMatrix matrixA = new SimpleMatrix(posList.size(), 2);
        SimpleMatrix vectorB = new SimpleMatrix(posList.size(), 1);

        while (error > MAX_DEVIATION && iterationIndex < MAX_ITERATIONS) {
            iterationIndex++;
            // Fill matrices with data
            for (int i = 0; i < posList.size(); i++) {
                GeoPosition pos = posList.get(i);
                double dis = distList.get(i);
                matrixA.set(i, 0, calcXPythDeriv(pos, estimation));
                matrixA.set(i, 1, calcYPythDeriv(pos, estimation));
                vectorB.set(i, 0, calcDeltaVecLength(pos, estimation, dis));
            }
            // Calculate result matrix
            SimpleMatrix matrixAT = matrixA.transpose();
            SimpleMatrix t1 = matrixAT.mult(matrixA);
            t1 = t1.invert();
            SimpleMatrix v1 = matrixAT.mult(vectorB);
            SimpleMatrix resultMatrix = t1.mult(v1);

            GeoPosition newEstimation = GeoPosition.fromCartesian(
                    resultMatrix.get(0) + estimation.getCartesianX(),
                    resultMatrix.get(1) + estimation.getCartesianY());

            error = getPosDistance(estimation, newEstimation);
            estimation = newEstimation;


            geoJson.add(String.format("{\"type\": \"Feature\", \n\"properties\": { \"label\": \"%d\", \"latlng\": \"%f, %f\"  }, \n\"geometry\": {\"type\": \"Point\", \"coordinates\": [%f, %f]} }\n", iterationIndex, estimation.getLng(), estimation.getLat(), estimation.getLng(), estimation.getLat()));
//			geoJson.add(String.format("{\"type\": \"Point\", \"coordinates\": [%f, %f], \"label\": %i }", estimation.getLng(), estimation.getLat(), iterationIndex));

            System.out.println("- Iteration: " + iterationIndex);
            System.out.println("--  Matrix A: " + matrixA.toString());
            System.out.println("-- Vector B: " + vectorB.toString());
            System.out.println("-- ErgebnisMatrix: " + resultMatrix.toString());
            System.out.println("- Estimation Error: " + error);
            System.out.println("----------");
            System.out.println("- Estimation Result: " + estimation.toString());
            System.out.println("----------");

        }
        return estimation;

    }

    /**
     * Calculates the center of the bounding box for all given points.
     *
     * @param points
     *            A list of points to calculate the bounding box from
     * @return A GeoPosition that represents the center of the bounding box of
     *         all given points
     */
    private GeoPosition getBoxCenter(ArrayList<GeoPosition> points) {
        double x_min = points.get(0).getCartesianX();
        double x_max = points.get(0).getCartesianX();
        double y_min = points.get(0).getCartesianY();
        double y_max = points.get(0).getCartesianY();
        for (int i = 1; i < points.size(); i++) {
            x_min = Math.min(x_min, points.get(i).getCartesianX());
            x_max = Math.max(x_max, points.get(i).getCartesianX());
            y_min = Math.min(y_min, points.get(i).getCartesianY());
            y_max = Math.max(y_max, points.get(i).getCartesianY());
        }
        return GeoPosition.fromCartesian((x_min + x_max) / 2,
                (y_min + y_max) / 2);
    }

    /**
     * Calculates the first pythagoras x derivation for the distance of two
     * points.
     *
     * @param p1
     *            First point
     * @param p2
     *            Second point
     * @return Resulting number
     */
    private double calcXPythDeriv(GeoPosition p1, GeoPosition p2) {
        double result = (-p1.getCartesianX() + p2.getCartesianX())
                / calcPythDerivDenom(p1, p2);
        return result;
    }

    /**
     * Calculates the first pythagoras y derivation for the distance of two
     * points.
     *
     * @param p1
     *            First point
     * @param p2
     *            Second point
     * @return Resulting number
     */
    private double calcYPythDeriv(GeoPosition p1, GeoPosition p2) {
        double result = (-p1.getCartesianY() + p2.getCartesianY())
                / calcPythDerivDenom(p1, p2);
        return result;
    }

    /**
     * Calculates the first pythagoras derivation denominator for the distance
     * of two points.
     *
     * @param p1
     *            First point
     * @param p2
     *            Second point
     * @return Resulting number
     */
    private double calcPythDerivDenom(GeoPosition p1, GeoPosition p2) {
        double result = Math.sqrt(
                Math.pow(p1.getCartesianX() - p2.getCartesianX(), 2)
                        + Math.pow(p1.getCartesianY() - p2.getCartesianY(), 2));
        return result;
    }

    /**
     * Calculates the length of the distance between vector length of two given
     * points and a given (distance-)length.
     *
     * @param p1
     *            First point
     * @param p2
     *            Second point
     * @param distance
     *            The distance to subtract the vector length from
     * @return Resulting length
     */
    private double calcDeltaVecLength(GeoPosition p1, GeoPosition p2,
                                      double distance) {
        return distance - calcPythDerivDenom(p1, p2);

    }

    /**
     * Calculates the distance between two points.
     *
     * @param p1
     *            First point
     * @param p2
     *            Second point
     * @return Resulting distance
     */
    private double getPosDistance(GeoPosition p1, GeoPosition p2) {
        return calcPythDerivDenom(p1, p2);
    }
}
