package de.uniba.kinf.geogames.esriguesstimate.data;

import uk.me.jstott.jcoord.ECEFRef;
import uk.me.jstott.jcoord.LatLng;

/**
 * Class that is used for geographic positions within the application.
 * GeoPositions can be constructed by latitude/longitude values or cartesian
 * coordinates that are projected through a Mercator projection.
 * 
 * see de.uniba.kinf.geogames.guesstimate.data.GeoPosition.fromCartesian
 * 
 * @author Thomas Heinz
 * 
 */
public class GeoPosition {

    private double lat;
    private double lng;
    private double x;
    private double y;
    private static double z;

    public static void init(double lat, double lng) {
        z = new ECEFRef(new LatLng(lat, lng)).getZ();
}

    // ===== Static Class Methods =====

    /**
     * Use a projection to transform the points latitude/longitude
     * values to cartesian coordinates.
     *
     * see https://en.wikipedia.org/wiki/Mercator_projection
     *
     * @param lat
     *            Latitude value for the projection (eg. 49.902833)
     * @param lng
     *            Longitude value for the projection (eg. 10.869787)
     * @return A double array that contains the cartesian x-value as its first
     *         entry. The cartesian y-value as its second entry.
     */
    public static double[] toCartesian(double lat, double lng) {
//        LatLng latlng = new LatLng(lat, lng);
//        OSRef osref = latlng.toOSRef();
//        double x = osref.getEasting();
//        double y = osref.getNorthing();
//        return new double[] { x, y };
        LatLng latlng = new LatLng(lat, lng);
        ECEFRef ref = new ECEFRef(latlng);
        return new double[] { ref.getX(), ref.getY() };
    }

    /**
     * Use a projection to transform the points cartesian coordinates
     * to latitude/longitude values.
     *
     * see https://en.wikipedia.org/wiki/Mercator_projection
     *
     * @param x
     *            The x-value for the projection.
     * @param y
     *            The y-value for the projection.
     * @return A double array that contains the latitude value as its first
     *         entry. The longitude y-value as its second entry.
     */
    public static double[] toLatLng(double x, double y) {
//        OSRef osref = new OSRef(x, y);
//        LatLng latLng = osref.toLatLng();
//        return new double[] { latLng.getLat(), latLng.getLng() };
        ECEFRef ref = new ECEFRef(x, y, z);
        LatLng latlng = ref.toLatLng();
        return new double[] { latlng.getLatitude(), latlng.getLongitude() };
    }

    /**
     * A static class method to create GeoPoints from cartesian coordinates.
     *
     * @param x
     *            Projected x-coordinate
     * @param y
     *            Projected y-coordinate
     * @return A GeoPosition with the corresponding properties
     */
    public static GeoPosition fromCartesian(double x, double y) {
        double[] latlng = toLatLng(x, y);
        GeoPosition pos = new GeoPosition();
        pos.x = x;
        pos.y = y;
        pos.lat = latlng[0];
        pos.lng = latlng[1];
        return pos;

    }

    public static GeoPosition fromGeoLocation(GeoLocation loc) {
        return new GeoPosition(loc.getLat(), loc.getLng());
    }

    // ===== Constructors =====

    /**
     * Constructor. For internal use only.
     */
    private GeoPosition() {
    };

    /**
     * Creates a new GeoPosition at the given latitude and longitude values.
     *
     * @param lat
     *            The latitude the position should be located at.
     * @param lng
     *            The longitude the position should be located at.
     */
    public GeoPosition(double lat, double lng) {
        this();
        this.lat = lat;
        this.lng = lng;
        // Calculate & set cartesian coordinates
        double[] cartesian = toCartesian(lat, lng);
        this.x = cartesian[0];
        this.y = cartesian[1];
    }

    /**
     * Returns the GeoPosition as a human readable string.
     *
     * @return Returns a string representation of the object.
     */
    @Override
    public String toString() {
        return "[GeoPosition (" + this.lat + ", " + this.lng + ") [" + this.x
                + ", " + this.y + "]";
    }

    // ===== Getter & Setter =====

    /**
     * Returns the GeoLocations latitude value.
     *
     * @return the GeoLocations latitude value as a double
     */
    public double getLat() {
        return lat;
    }

    /**
     * Returns the GeoLocations longitude value.
     *
     * @return the GeoLocations longitude value as a double
     */
    public double getLng() {
        return lng;
    }

    /**
     * Returns the GeoLocations latitude value in E6 format.
     *
     * @return the GeoLocations latitude value as an integer in the E6 format.
     *         MicroDegrees (the E6 format) = DecimalDegrees * 1e6
     */
    public int getLatE6() {
        return (int) Math.round(this.lat * 1E6);
    }

    /**
     * Returns the GeoLocations longitude value in E6 format.
     *
     * @return the GeoLocations longitude value as an integer in the E6 format.
     *         MicroDegrees (the E6 format) = DecimalDegrees * 1e6
     */
    public int getLngE6() {
        return (int) Math.round(this.lng * 1E6);
    }

    /**
     * Returns the GeoLocations x-position as a Cartesian coordinate.
     *
     * see toCartesian
     *
     * @return Returns the GeoPoints Cartesian x coordinate.
     */
    public double getCartesianX() {
        return this.x;
    }

    /**
     * Returns the GeoLocations y-position as a Cartesian coordinate.
     *
     * see toCartesian
     *
     * @return Returns the GeoPoints Cartesian y coordinate.
     */
    public double getCartesianY() {
        return this.y;
    }
//
//	/**
//	 * Approximated earth radius in meters.
//	 */
//	private static final int EARTH_RADIUS = 6378137;
//
////	private static CoordinateReferenceSystem wgs84Crs;
////	private static CoordinateReferenceSystem utmCrs;
//
//	private double lat;
//	private double lng;
//	private double x;
//	private double y;
//
//	// ===== Static Class Methods =====
//
//	/**
//	 * Use the Mercator projection to transform the points latitude/longitude
//	 * values to cartesian coordinates.
//	 *
//	 * see https://en.wikipedia.org/wiki/Mercator_projection
//	 *
//	 * @param lat
//	 *            Latitude value for the projection (eg. 49.902833)
//	 * @param lng
//	 *            Longitude value for the projection (eg. 10.869787)
//	 * @return A double array that contains the cartesian x-value as its first
//	 *         entry. The cartesian y-value as its second entry.
//	 */
//	public static double[] toCartesian(double lat, double lng) {
//		// Use JCOORD
//		LatLng latlng = new LatLng(lat, lng);
//		OSRef osref = latlng.toOSRef();
//		double x = osref.getEasting();
//		double y = osref.getNorthing();
//		return new double[] { x, y };
////		double x = Math.toRadians(lng) * EARTH_RADIUS;
////		double y = EARTH_RADIUS
////				* Math.log(Math.tan(Math.PI / 4 + Math.toRadians(lat) / 2));
////		return new double[] { x, y };
//
//
//
////		MathTransform mathTransform = CRS.findMathTransform(wgs84Crs,
////				utmCrs, true);
////		DirectPosition2D srcDirectPosition2D = new DirectPosition2D(
////				wgs84Crs, lat, lng);
////		DirectPosition2D destDirectPosition2D = new DirectPosition2D();
////		mathTransform.transform(srcDirectPosition2D, destDirectPosition2D);
////		mathTransform.t
//	}
//
//	/**
//	 * Use the Mercator projection to transform the points cartesian coordinates
//	 * to latitude/longitude values.
//	 *
//	 * see https://en.wikipedia.org/wiki/Mercator_projection
//	 *
//	 * @param x
//	 *            The x-value for the projection.
//	 * @param y
//	 *            The y-value for the projection.
//	 * @return A double array that contains the latitude value as its first
//	 *         entry. The longitude y-value as its second entry.
//	 */
//	public static double[] toLatLng(double x, double y) {
//		OSRef osref = new OSRef(x, y);
//		LatLng latLng = osref.toLatLng();
////		double lat = Math.toDegrees(2 * Math.atan(Math.exp(y / EARTH_RADIUS))
////				- Math.PI / 2);
////		double lng = Math.toDegrees(x / EARTH_RADIUS);
//		return new double[] { latLng.getLat(), latLng.getLng() };
//	}
//
//	/**
//	 * A static class method to create GeoPoints from cartesian coordinates.
//	 *
//	 * @param x
//	 *            Projected x-coordinate
//	 * @param y
//	 *            Projected y-coordinate
//	 * @return A GeoPosition with the corresponding properties
//	 */
//	public static GeoPosition fromCartesian(double x, double y) {
//		double[] latlng = toLatLng(x, y);
//		GeoPosition pos = new GeoPosition();
//		pos.x = x;
//		pos.y = y;
//		pos.lat = latlng[0];
//		pos.lng = latlng[1];
//		return pos;
//	}
//
//    public static GeoPosition fromGeoLocation(GeoLocation loc) {
//        return new GeoPosition(loc.getLat(), loc.getLng());
//    }
//
//	// ===== Constructors =====
//
//	/**
//	 * Constructor. For internal use only.
//	 */
//	private GeoPosition() {
//	};
//
//	/**
//	 * Create a GeoPosition from a given GeoPoint.
//	 *
//	 * @param geoPoint
//	 *            GeoPoint which's location is read to create the GeoPosition.
//	 */
////	public GeoPosition(GeoPoint geoPoint) {
////		this(geoPoint.getLatitudeE6() / 1E6, geoPoint.getLongitudeE6() / 1E6);
////	}
//
//	/**
//	 * Creates a new GeoPosition at the given latitude and longitude values.
//	 *
//	 * @param lat
//	 *            The latitude the position should be located at.
//	 * @param lng
//	 *            The longitude the position should be located at.
//	 */
//	public GeoPosition(double lat, double lng) {
//		this();
//		this.lat = lat;
//		this.lng = lng;
//		// Calculate & set cartesian coordinates
//		double[] cartesian = toCartesian(lat, lng);
//		this.x = cartesian[0];
//		this.y = cartesian[1];
//	}
//
//	/**
//	 * Returns the GeoPosition as a human readable string.
//	 *
//	 * @return Returns a string representation of the object.
//	 */
//	@Override
//	public String toString() {
//		return "[GeoPosition (" + this.lat + ", " + this.lng + ") [" + this.x
//				+ ", " + this.y + "]";
//	}
//
//	// ===== Getter & Setter =====
//
//	/**
//	 * Returns the GeoLocations latitude value.
//	 *
//	 * @return the GeoLocations latitude value as a double
//	 */
//	public double getLat() {
//		return lat;
//	}
//
//	/**
//	 * Returns the GeoLocations longitude value.
//	 *
//	 * @return the GeoLocations longitude value as a double
//	 */
//	public double getLng() {
//		return lng;
//	}
//
//	/**
//	 * Returns the GeoLocations latitude value in E6 format.
//	 *
//	 * @return the GeoLocations latitude value as an integer in the E6 format.
//	 *         MicroDegrees (the E6 format) = DecimalDegrees * 1e6
//	 */
//	public int getLatE6() {
//		return (int) Math.round(this.lat * 1E6);
//	}
//
//	/**
//	 * Returns the GeoLocations longitude value in E6 format.
//	 *
//	 * @return the GeoLocations longitude value as an integer in the E6 format.
//	 *         MicroDegrees (the E6 format) = DecimalDegrees * 1e6
//	 */
//	public int getLngE6() {
//		return (int) Math.round(this.lng * 1E6);
//	}
//
//	/**
//	 * Returns the GeoLocations x-position as a Cartesian coordinate.
//	 *
//	 * see toCartesian
//	 *
//	 * @return Returns the GeoPoints Cartesian x coordinate.
//	 */
//	public double getCartesianX() {
//		return this.x;
//	}
//
//	/**
//	 * Returns the GeoLocations y-position as a Cartesian coordinate.
//	 *
//	 * see toCartesian
//	 *
//	 * @return Returns the GeoPoints Cartesian y coordinate.
//	 */
//	public double getCartesianY() {
//		return this.y;
//	}
//


}
